import logging
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Protocol

from iterpy import Iter

from .document import Document
from .document_source import BaseDocumentSource
from .extractors.extractor import BasePromptExtractor
from .prompts.prompt import BasePrompt

log = logging.getLogger(__name__)


class BasePromptSource(Protocol):
    def get_prompts(self) -> Sequence[BasePrompt]: ...


@dataclass(frozen=True)
class DocumentPromptSource(BasePromptSource):
    document_ingester: BaseDocumentSource
    prompt_extractors: Sequence[BasePromptExtractor]

    def _get_prompts_from_document(self, document: Document) -> Sequence[BasePrompt]:
        prompts: list[BasePrompt] = []

        for extractor in self.prompt_extractors:
            try:
                extractor_prompts = list(extractor.extract_prompts(document))
                prompts += extractor_prompts
            except Exception as e:
                log.error(
                    f"Failed to extract prompts with {extractor} from {document.source_path.name} using {extractor}. Reason: {e}"
                )

        return prompts

    def _deduplicate_group(self, group: tuple[str, Sequence[BasePrompt]]) -> BasePrompt:
        prompts_in_group = group[1]

        if len(prompts_in_group) != 1:
            identifier = (
                prompts_in_group[0].edit_url
                if prompts_in_group[0].edit_url
                else prompts_in_group[0]
            )
            log.warning(f"""Duplicate prompts: {identifier}""")

        return prompts_in_group[0]

    def _deduplicate_prompts(self, prompts: Sequence[BasePrompt]) -> Sequence[BasePrompt]:
        """Deduplicate prompts based on scheduling UID. If the scheduling UID is the same, the prompt is considered a duplicate."""
        scheduling_uuid_groups = Iter(prompts).groupby(lambda prompt: str(prompt.scheduling_uid))

        unique_prompts = scheduling_uuid_groups.map(self._deduplicate_group).to_list()

        n_duplicates = len(prompts) - len(unique_prompts)
        if n_duplicates != 0:
            log.warning(f"Found a total of {n_duplicates} duplicate prompts")

        return unique_prompts

    def get_prompts(self) -> Sequence[BasePrompt]:
        prompts = (
            Iter(self.document_ingester.get_documents())
            .map(self._get_prompts_from_document)
            .flatten()
            .to_list()
        )

        return self._deduplicate_prompts(prompts)



from collections.abc import Sequence
from dataclasses import dataclass

from ...utils.hash_cleaned_str import clean_str, hash_str_to_int
from .prompt import BasePrompt
from .prompt_from_doc import PromptFromDocMixin


@dataclass(frozen=True)
class ClozePrompt(BasePrompt):
    text: str

    @property
    def scheduling_uid(self) -> int:
        return hash_str_to_int(clean_str(self.text))

    @property
    def update_uid(self) -> int:
        return hash_str_to_int(f"{(self.text)}{self.tags}")

    @property
    def tags(self) -> Sequence[str]:
        return ()


@dataclass(frozen=True)
class ClozeWithoutDoc(ClozePrompt):
    add_tags: Sequence[str]

    @property
    def tags(self) -> Sequence[str]:
        return self.add_tags

    @property
    def edit_url(self) -> str | None:
        return None


@dataclass(frozen=True)
class ClozeFromDoc(ClozePrompt, PromptFromDocMixin):
    text: str

    @property
    def tags(self) -> Sequence[str]:
        return self.parent_doc.tags



from collections.abc import Sequence
from dataclasses import dataclass


@dataclass(frozen=True)
class LineBlock:
    starting_line: int
    lines: Sequence[str]

    @property
    def content(self) -> str:
        return "\n".join(self.lines)

    @property
    def end_line(self) -> int:
        return self.starting_line + len(self.lines) - 1


def to_line_blocks(content: str) -> Sequence[LineBlock]:
    lines = content.split("\n")

    line_blocks: list[LineBlock] = []
    cur_lines: Sequence[str] = []

    cur_starting_line = 0
    for i, line in enumerate(lines):
        cur_is_newline = line == ""  # Since we split on \n, we can check for empty lines
        if cur_is_newline:
            line_blocks.append(LineBlock(starting_line=cur_starting_line, lines=cur_lines))
            cur_starting_line = i + 1
            cur_lines = []
        else:
            cur_lines.append(line)

        final_line = i == len(lines) - 1
        if final_line:
            line_blocks.append(LineBlock(starting_line=cur_starting_line, lines=cur_lines))

    return line_blocks



import logging
import warnings
from collections.abc import Sequence

import genanki
from iterpy import Iter

from ..source.prompts.prompt import DestinationPrompt
from ..utils.hash_cleaned_str import clean_str, hash_str_to_int
from .ankiconnect.anki_converter import AnkiPromptConverter
from .ankiconnect.anki_prompt import AnkiPrompt
from .ankiconnect.ankiconnect_gateway import AnkiConnectGateway
from .destination import DeletePrompts, PromptDestination, PromptDestinationCommand, PushPrompts

log = logging.getLogger(__name__)


class AnkiConnectDestination(PromptDestination):
    def __init__(self, gateway: AnkiConnectGateway, prompt_converter: AnkiPromptConverter) -> None:
        self.gateway = gateway
        self.prompt_converter = prompt_converter

        # Don't care about genanki warnings, have our own tests
        warnings.filterwarnings(
            "ignore", module="genanki", message="^Field contained the following invalid HTML tags"
        )

    def get_all_prompts(self) -> Sequence[DestinationPrompt]:
        return (
            Iter(self.gateway.get_all_note_infos())
            .map(self.prompt_converter.note_info_to_prompt)
            .to_list()
        )

    def _delete_prompts(self, prompts: Sequence[DestinationPrompt]) -> None:
        prompt_ids = {int(remote_prompt.destination_id) for remote_prompt in prompts}
        log.info(f"Deleting {len(prompt_ids)} prompts from Anki")
        for prompt in prompts:
            log.info(f"Deleting prompt: {prompt.prompt.edit_url}")

        self.gateway.delete_notes(list(prompt_ids))

    def _grouped_cards_to_deck(
        self, grouped_cards: tuple[str, Sequence[AnkiPrompt]]
    ) -> genanki.Deck:
        deck_name = grouped_cards[0]
        deck = genanki.Deck(name=deck_name, deck_id=hash_str_to_int(clean_str(deck_name)))

        for card in grouped_cards[1]:
            deck.add_note(card.to_genanki_note())  # type: ignore

        return deck

    def _create_package(self, cards: Sequence[AnkiPrompt]) -> genanki.Package:
        decks = (
            Iter(cards).groupby(lambda card: card.deck).map(self._grouped_cards_to_deck).to_list()
        )

        return genanki.Package(deck_or_decks=decks)

    def _push_prompts(self, command: PushPrompts) -> None:
        cards = [self.prompt_converter.prompt_to_card(e) for e in command.prompts]

        models = [card.genanki_model for card in cards]
        unique_models: dict[int, genanki.Model] = {
            model.model_id: model  # type: ignore
            for model in models  # type: ignore
        }

        for model in unique_models.values():
            self.gateway.update_model(model)

        package = self._create_package(cards)

        log.info(f"Pushing {len(cards)} cards to Anki")
        self.gateway.import_package(package)

    def update(self, commands: Sequence[PromptDestinationCommand]) -> None:
        for command in commands:
            match command:
                case DeletePrompts(prompts):
                    self._delete_prompts(prompts)
                case PushPrompts(prompts):
                    self._push_prompts(command)



from collections.abc import Sequence
from dataclasses import dataclass

import pytest

from memium.utils.extract_terms import get_terms_surrounded_by_underscores


@dataclass(frozen=True)
class Example:
    input_string: str
    wikilinks: Sequence[str]


@pytest.mark.parametrize(
    ("example"),
    [
        Example("_Test_", ["Test"]),
        Example("_Test (testing)_", ["Test (testing)"]),
        Example("_Test_, _Test2_", ["Test", "Test2"]),
        Example("_ Some things, are just not _Wikilinks_ _", ["Wikilinks"]),
    ],
    ids=lambda x: x.input_string,
)
def test_get_wikilinks_from_string(example: Example):
    assert get_terms_surrounded_by_underscores(example.input_string) == example.wikilinks
